#include "ImageRoutinen.h"
#include "IMImage.h"
#include "list" 
using namespace std; 

// ------------------------------------ Plugin Basis -----------------------------------

long * IMObjectsarray;
long IMObjectsarraySize;

void InitPlugin(void)
{
	IMObjectsarray = (long *) malloc(0);
	IMObjectsarraySize = 0;

	InitializeMagick(NULL);  
}

void DeinitPlugin(void)
{
	free(IMObjectsarray);
	//  DestroyMagick();   // unknown?
}

// ------------------------------------ Convert -----------------------------------


void IM_GetLastError( PA_PluginParameters params)
{
	char result[255];

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		strlcpy(result,  newImage->GetLastError(), 255);
	}
	else
		result[0] = 0;  // unkown object

	PA_ReturnString( params, result );
}


void IM_Open( PA_PluginParameters params )
{
	long Path_len;
	char* Path;
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		Path_len = PA_GetTextParameter( params, 2, 0L );	
		Path = (char *) malloc(Path_len+1);
		Path_len = PA_GetTextParameter( params, 2, Path );
		Path[Path_len] = 0;
		
#if VERSIONWIN
		returnValue = newImage->read(Path);
#else			
		Path_len= Path_len * 3;
		UInt8		*NewPath_utf8   = NULL;
		NewPath_utf8 = (UInt8*) malloc ( Path_len );  // to have enough space for diacritical chars
		if ( NewPath_utf8 != NULL ) 
			returnValue = build_4DPath_2_PosixPath (Path , 0L , NewPath_utf8 , Path_len );
		else
			returnValue = kErr_MemError;
		
		if (( NewPath_utf8 != NULL ) && ( returnValue == noErr ))
			returnValue = newImage->read((char *) NewPath_utf8);
		
		if ( NewPath_utf8 != NULL )
			free ( NewPath_utf8);
#endif
		free(Path);
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );

}


void IM_OpenBlob( PA_PluginParameters params )
{
	long Blob_len;
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		PA_Handle myblob = PA_GetBlobHandleParameter( params, 2 );	
		Blob_len = PA_GetHandleSize(myblob);
		
		returnValue = newImage->read(PA_LockHandle(myblob), Blob_len);
		PA_UnlockHandle(myblob);
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );
}


void IM_Save( PA_PluginParameters params )
{
	long Path_len;
	char* Path;
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		Path_len = PA_GetTextParameter( params, 2, 0L );	
		Path = (char *) malloc(Path_len+1);
		Path_len = PA_GetTextParameter( params, 2, Path );
		Path[Path_len] = 0;
		
#if VERSIONWIN
		returnValue = newImage->write(Path);
#else			
		Path_len= Path_len * 3;
		UInt8		*NewPath_utf8   = NULL;
		NewPath_utf8 = (UInt8*) malloc ( Path_len );  // to have enough space for diacritical chars
		if ( NewPath_utf8 != NULL ) 
			returnValue = build_4DPath_2_PosixPath (Path , 0L , NewPath_utf8 , Path_len );
		else
			returnValue = kErr_MemError;
		
		if (( NewPath_utf8 != NULL ) && ( returnValue == noErr ))
			returnValue = newImage->write((char *) NewPath_utf8);
		
		if ( NewPath_utf8 != NULL )
			free ( NewPath_utf8);
#endif

		free(Path);
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );
}

void IM_SaveBlob( PA_PluginParameters params )
{
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		PA_Handle blob = PA_NewHandle(0);
		returnValue = newImage->write(blob);
		PA_SetBlobHandleParameter(params, 2, blob);
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );
}

void IM_GetImageProperty( PA_PluginParameters params)
{
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		double para1, para2, para3, para4; 
		char stringpara[255];
		para1 = para2 = para3 = para4 = 0;

		long Selector = PA_GetLongParameter( params, 2 );
		PA_GetStringParameter(params, 3, stringpara);
		returnValue = newImage->GetProperty(Selector, stringpara, para1, para2, para3, para4);
		PA_SetStringParameter(params, 3, stringpara);
		PA_SetDoubleParameter(params, 4, para1);
		PA_SetDoubleParameter(params, 5, para2);
		PA_SetDoubleParameter(params, 6, para3);
		PA_SetDoubleParameter(params, 7, para4);
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );
}

void IM_SetImageProperty( PA_PluginParameters params)
{
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		double para1, para2, para3, para4; 
		char stringpara[255];
		para1 = para2 = para3 = para4 = 0;
		stringpara[0] = 0;

		long Selector = PA_GetLongParameter( params, 2 );
		PA_GetStringParameter(params, 3, stringpara);
		para1 = PA_GetDoubleParameter(params, 4);
		para2 = PA_GetDoubleParameter(params, 5);
		para3 = PA_GetDoubleParameter(params, 6);
		para4 = PA_GetDoubleParameter(params, 7);
		returnValue = newImage->SetProperty(Selector, stringpara, para1, para2, para3, para4);
		
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );
}


void IM_ModifyImage( PA_PluginParameters params)
{
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		double para1, para2, para3, para4, para5, para6; 
		char stringpara[255];
		para1 = para2 = para3 = para4 = para5 = para6 = 0;
		stringpara[0] = 0;

		long Selector = PA_GetLongParameter( params, 2 );
		PA_GetStringParameter(params, 3, stringpara);
		para1 = PA_GetDoubleParameter(params, 4);
		para2 = PA_GetDoubleParameter(params, 5);
		para3 = PA_GetDoubleParameter(params, 6);
		para4 = PA_GetDoubleParameter(params, 7);
		para5 = PA_GetDoubleParameter(params, 8);
		para6 = PA_GetDoubleParameter(params, 9);
		returnValue = newImage->ModifyImage(Selector, stringpara, para1, para2, para3, para4, para5, para6);	
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );
}


void IM_New(PA_PluginParameters params )
{
	IMImage * newImage;

	newImage = new(IMImage);

	// add entry to our array
	IMObjectsarraySize++;
	IMObjectsarray = (long *) realloc(IMObjectsarray, IMObjectsarraySize * sizeof(long));
	IMObjectsarray[IMObjectsarraySize-1] = (long) newImage;

	PA_ReturnLong( params, (long) newImage );
}

void IM_Clear(PA_PluginParameters params )
{
	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		// remove entry from our array
		IMObjectsarray[CurrentPos] = IMObjectsarray[IMObjectsarraySize-1];
		IMObjectsarraySize--;
		IMObjectsarray = (long *) realloc(IMObjectsarray, IMObjectsarraySize * sizeof(long));

		free(newImage);
		PA_ReturnLong( params, 0 );
	}
	else
		PA_ReturnLong( params, ImageMagickErrorUnknownObject );  // unkown object
}

void IM_Copy( PA_PluginParameters params )
{
	long returnValue = 0;

	IMImage * newImage;
	IMImage * oldImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	oldImage = (IMImage *) PA_GetLongParameter( params, 2 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		CurrentPos = ValidateIMObject((long) oldImage);
		if (CurrentPos >=0)
			*newImage = *oldImage;
		else
			returnValue = ImageMagickErrorUnknownObject;  // unkown object
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );

}

void IM_Create( PA_PluginParameters params )
{
	long returnValue = 0;
	char stringpara[255];

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		int x = PA_GetLongParameter( params, 2 );	
		int y = PA_GetLongParameter( params, 3 );	
		PA_GetStringParameter(params, 4, stringpara);
		
		returnValue = newImage->create(x, y, stringpara);
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );

}

void IM_Draw( PA_PluginParameters params)
{
	long returnValue = 0;

	IMImage * newImage;
	newImage = (IMImage *) PA_GetLongParameter( params, 1 );	
	long CurrentPos = ValidateIMObject((long) newImage);
	if (CurrentPos >=0)
	{	
		double para1, para2, para3, para4, para5, para6; 
		char stringpara[255];
		para1 = para2 = para3 = para4 = para5 = para6 = 0;
		stringpara[0] = 0;

		long Selector = PA_GetLongParameter( params, 2 );
		PA_GetStringParameter(params, 3, stringpara);
		para1 = PA_GetDoubleParameter(params, 4);
		para2 = PA_GetDoubleParameter(params, 5);
		para3 = PA_GetDoubleParameter(params, 6);
		para4 = PA_GetDoubleParameter(params, 7);
		para5 = PA_GetDoubleParameter(params, 8);
		para6 = PA_GetDoubleParameter(params, 9);
		returnValue = newImage->Draw(Selector, stringpara, para1, para2, para3, para4, para5, para6);
		
	}
	else
		returnValue = ImageMagickErrorUnknownObject;  // unkown object

	PA_ReturnLong( params, returnValue );
}



// ****************************************************************************************

void IM_GetSupportedFormats(PA_PluginParameters params )
{
  long returnValue = 0;
  string Formatdesc;
  list<CoderInfo> coderList;
  coderInfoList( &coderList,           // Reference to output list
                 CoderInfo::TrueMatch, // Match readable formats
                 CoderInfo::AnyMatch,  // Don't care about writable formats
                 CoderInfo::AnyMatch); // Don't care about multi-frame support
  list<CoderInfo>::iterator entry = coderList.begin();

  PA_Variable Tag = PA_GetVariableParameter( params, 1 );
  if (Tag.fType != eVK_ArrayText)
	returnValue = ImageMagickErrorUnknownCommand;
  else
  {
	  PA_ResizeArray (&Tag, 0);
	  long loopcounter = 0;

	  while( entry != coderList.end() )
	  {
		Formatdesc = entry->name() + " (" + entry->description() + ") ";
		if ( entry->isReadable() )
		  Formatdesc += "r";
		else
		  Formatdesc += " ";
		if ( entry->isWritable() )
		  Formatdesc += "w";
		else
		  Formatdesc += " ";
		if ( entry->isMultiFrame() )
		  Formatdesc += "m";
		else
		  Formatdesc += " ";

		  PA_ResizeArray (&Tag, ++loopcounter);
		  PA_SetTextInArray (Tag, loopcounter, (char *) Formatdesc.c_str(), Formatdesc.length());

		entry ++;
	   } 
	  PA_SetVariableParameter( params, 1, Tag, 0 );
  }

  PA_ReturnLong( params, returnValue );
}


// ------------------------------------ Helpers -----------------------------------

long ValidateIMObject(long newImage)
{	// test if it is a valid object/pointer
	for (long i = 0; i<IMObjectsarraySize; i++)
	{
		if (IMObjectsarray[i] == newImage)
			return i;
	}
	return ImageMagickErrorUnknownObject;
}


/*
 INTERNAL USE - MAC ONLY
 We are able to open a fullpath > 255, and unicode.
 */
#if VERSIONMAC
// NewPath		can be NULL		Encoding MacRoman
// newPathUTF8  can be NULL		Encoding UTF8
OSErr build_4DPath_2_PosixPath (char* Pfad , char *NewPath , UInt8 *newPathUTF8 , CFIndex buffSize  )
{
	OSErr		err			= noErr;
	CFStringRef filePath	= NULL;
	CFURLRef	url			= NULL;
	CFStringRef altpath		= NULL ;
	TextEncoding			myTextEncoding = kCFStringEncodingMacRoman;  
	
	filePath = CFStringCreateWithBytes ( kCFAllocatorDefault,
										 (const UInt8*) Pfad,
										 (CFIndex) strlen(Pfad),
										 myTextEncoding, /* CFStringEncoding encoding,*/
										 false);
	
	if ( filePath != NULL )
		{
		url = CFURLCreateWithFileSystemPath(NULL, filePath, kCFURLHFSPathStyle, false);
		if ( url != NULL )
			{
			altpath = CFURLCopyFileSystemPath(url, kCFURLPOSIXPathStyle);
			if ( altpath != NULL )
				{
				if ( newPathUTF8 != NULL)
					CFStringGetCString (altpath, (char*)newPathUTF8, buffSize, kCFStringEncodingUTF8 );
				
				if ( NewPath != NULL)
					CFStringGetCString (altpath, NewPath, buffSize, kCFStringEncodingMacRoman );				
				}
			}
		}
	
	if ( filePath != NULL )
		CFRelease(filePath);
	else
		err = kErr_MemError;
	
	if ( url != NULL )
		CFRelease(url);
	else
		err = kErr_MemError;
	
	if ( altpath != NULL )
		CFRelease(altpath);
	else
		err = kErr_MemError;
	
	return ( err );
	
}
#endif